#!/usr/bin/python3.8
import os
import logging
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib
import conf2Utils
import psutil
from configparser import ConfigParser

from collections import Sequence
import _dbus_bindings
from dbus import (
    INTROSPECTABLE_IFACE, ObjectPath, SessionBus, Signature, Struct,
    validate_bus_name, validate_object_path)
from dbus.decorators import method, signal
from dbus.exceptions import (
    DBusException, NameExistsException, UnknownMethodException)
from dbus.lowlevel import ErrorMessage, MethodReturnMessage, MethodCallMessage
from dbus.proxies import LOCAL_PATH
from dbus._compat import is_py2

logger = None

# 创建一个DBus服务类
class KYSettingsDBusService(dbus.service.Object):
    def __init__(self, bus_name, object_path):
        logger.info(f'DBus Class init')
        dbus.service.Object.__init__(self, bus_name, object_path)

        self._converter = conf2Utils.Converter(logger)

        self._reload()
        
        if os.path.exists('/etc/dbus-1/conf/com.kylin.kysdk.conf2.limit'):
            conf = ConfigParser()
            conf.read('/etc/dbus-1/conf/com.kylin.kysdk.conf2.limit')
            # conf.read('/home/kylin/disk/kylinos-src/libkysdk-base/src/conf2/service/com.kylin.kysdk.conf2.limit')
            if conf.has_section('whitelist'):
                self._whitelist =  list(conf['whitelist'].values())

        logger.info(f'DBus Class init end')
    
    def _is_allowed_caller(self, path):
        try:
            if not hasattr(self, '_whitelist'):
                return True
            return path in self._whitelist
        except:
            return False
    
    def _message_cb(self, connection, message):
        if not isinstance(message, MethodCallMessage):
            return

        try:
            sender = message.get_sender()
            bus = dbus.SystemBus()
            obj = bus.get_object('org.freedesktop.DBus', '/org/freedesktop/DBus')
            interface = dbus.Interface(obj, 'org.freedesktop.DBus')
            process = psutil.Process(interface.GetConnectionUnixProcessID(sender))
            executable_path = process.exe()
            if executable_path == '/usr/bin/python3.8':
                executable_path = process.cmdline()[1]
            if False == self._is_allowed_caller(executable_path):
                raise PermissionError('dbus limit control,not supported')
            
            # lookup candidate method and parent method
            method_name = message.get_member()
            interface_name = message.get_interface()
            (candidate_method, parent_method) = dbus.service._method_lookup(self, method_name, interface_name)

            # set up method call parameters
            args = message.get_args_list(**parent_method._dbus_get_args_options)
            keywords = {}

            if parent_method._dbus_out_signature is not None:
                signature = Signature(parent_method._dbus_out_signature)
            else:
                signature = None

            # set up async callback functions
            if parent_method._dbus_async_callbacks:
                (return_callback, error_callback) = parent_method._dbus_async_callbacks
                keywords[return_callback] = lambda *retval: dbus.service._method_reply_return(connection, message, method_name, signature, *retval)
                keywords[error_callback] = lambda exception: dbus.service._method_reply_error(connection, message, exception)

            # include the sender etc. if desired
            if parent_method._dbus_sender_keyword:
                keywords[parent_method._dbus_sender_keyword] = message.get_sender()
            if parent_method._dbus_path_keyword:
                keywords[parent_method._dbus_path_keyword] = message.get_path()
            if parent_method._dbus_rel_path_keyword:
                path = message.get_path()
                rel_path = path
                for exp in self._locations:
                    # pathological case: if we're exported in two places,
                    # one of which is a subtree of the other, then pick the
                    # subtree by preference (i.e. minimize the length of
                    # rel_path)
                    if exp[0] is connection:
                        if path == exp[1]:
                            rel_path = '/'
                            break
                        if exp[1] == '/':
                            # we already have rel_path == path at the beginning
                            continue
                        if path.startswith(exp[1] + '/'):
                            # yes we're in this exported subtree
                            suffix = path[len(exp[1]):]
                            if len(suffix) < len(rel_path):
                                rel_path = suffix
                rel_path = ObjectPath(rel_path)
                keywords[parent_method._dbus_rel_path_keyword] = rel_path

            if parent_method._dbus_destination_keyword:
                keywords[parent_method._dbus_destination_keyword] = message.get_destination()
            if parent_method._dbus_message_keyword:
                keywords[parent_method._dbus_message_keyword] = message
            if parent_method._dbus_connection_keyword:
                keywords[parent_method._dbus_connection_keyword] = connection

            # call method
            retval = candidate_method(self, *args, **keywords)

            # we're done - the method has got callback functions to reply with
            if parent_method._dbus_async_callbacks:
                return

            # otherwise we send the return values in a reply. if we have a
            # signature, use it to turn the return value into a tuple as
            # appropriate
            if signature is not None:
                signature_tuple = tuple(signature)
                # if we have zero or one return values we want make a tuple
                # for the _method_reply_return function, otherwise we need
                # to check we're passing it a sequence
                if len(signature_tuple) == 0:
                    if retval == None:
                        retval = ()
                    else:
                        raise TypeError('%s has an empty output signature but did not return None' %
                            method_name)
                elif len(signature_tuple) == 1:
                    retval = (retval,)
                else:
                    if isinstance(retval, Sequence):
                        # multi-value signature, multi-value return... proceed
                        # unchanged
                        pass
                    else:
                        raise TypeError('%s has multiple output values in signature %s but did not return a sequence' %
                            (method_name, signature))

            # no signature, so just turn the return into a tuple and send it as normal
            else:
                if retval is None:
                    retval = ()
                elif (isinstance(retval, tuple)
                      and not isinstance(retval, Struct)):
                # If the return is a tuple that is not a Struct, we use it
                # as-is on the assumption that there are multiple return
                # values - this is the usual Python idiom. (fd.o #10174)
                    pass
                else:
                    retval = (retval,)

            dbus.service._method_reply_return(connection, message, method_name, signature, *retval)
        except Exception as exception:
            # send error reply
            dbus.service._method_reply_error(connection, message, exception)

    # 重载
    def _reload(self):
        if self._converter.update_read_only_db():
            if self._converter.update_user_db('/root/.config/kylin-config/user.db'):
                self.updated()
                return True
        return False

    # key值改变信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='sss')
    def key_changed(self, id, version, key):
        pass

    # 配置信息update信号
    @dbus.service.signal("com.kylin.kysdk.conf2", signature='')
    def updated(self):
        pass

    # 设置一个配置的值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='sssss', out_signature='i')
    def set(self, user, id, version, key, value):
        success = False
        if user == 'root':
            success = self._converter.set('/root/.config/kylin-config/user.db',  id, version, key, value)
        else:
            success = self._converter.set(f'/home/{user}/.config/kylin-config/user.db', id, version, key, value)

        if success:
            self.key_changed(id, version, key)

        return success
    
    # 恢复单个默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ssss', out_signature='i')
    def reset(self,user, id, version, key):
        success = False
        if user == 'root':
            success = self._converter.reset('/root/.config/kylin-config/user.db',  id, version, key)
        else:
            success = self._converter.reset(f'/home/{user}/.config/kylin-config/user.db',  id, version, key)

        if success:
            self.key_changed(id, version, key)
        
        return success

    # 恢复整个id的默认值
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='', out_signature='i')
    def reload(self):
        success = self._reload()
        self.updated if success else None
        return success

    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ss', out_signature='b')
    def save(self, user, path):
        success = False
        if user == 'root':
            success = self._converter.save('/root/.config/kylin-config/user.db', path)
        else:
            success = self._converter.save(f'/home/{user}/.config/kylin-config/user.db', path)
        return success

    #复制old_id的配置到new_id的位置，并将新的数据写入到数据库
    @dbus.service.method("com.kylin.kysdk.conf2", in_signature='ssss', out_signature='b')
    def extends_id(self, old_id:str, new_id:str, version:str, user:str):
        success = False
        if user == 'root':
            success = self._converter.extends_id(old_id,  new_id, version, '/root/.config/kylin-config/user.db')
        else:
            success = self._converter.extends_id(old_id,  new_id, version, f'/home/{user}/.config/kylin-config/user.db')
        return success

if __name__ == "__main__":
    os.makedirs('/etc/kylin-config', exist_ok=True)
    logging.basicConfig(filename='/etc/kylin-config/systembus.log', 
                        level=logging.DEBUG, 
                        format='%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s', 
                        datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger()

    # 初始化DBus主循环
    DBusGMainLoop(set_as_default=True)

    # 连接到DBus系统总线
    bus = dbus.SystemBus()

    # 注册一个DBus服务
    bus_name = dbus.service.BusName("com.kylin.kysdk.conf2", bus=bus)
    object_path = "/com/kylin/kysdk/conf2"
    my_dbus_service = KYSettingsDBusService(bus_name, object_path)

    # 开始DBus主循环
    mainloop = GLib.MainLoop()
    mainloop.run()